#include "single_var.h"
namespace single_var
{
    int read_configfile( 
        const std::string & config_filename,
              std::string & geno_filename,
              std::string & out_filename,
                   size_t & start_rep,
                   size_t & num_reps,     
                   double & MAF_cutoff,
                 bool & use_pheno,
                 std::string & pheno_filename,
                   size_t & out_freq,
                   double & min_call_rate  )
    {
       bool got_geno_filename = false;
       bool got_out_filename = false;
       bool got_start_rep = false;
       bool got_num_reps = false;
       bool got_MAF_cutoff = false;
       const double DEF_MIN_CALL_RATE = 0.90;
       min_call_rate = DEF_MIN_CALL_RATE;
       size_t min_MAC = 0;
       min_MAC = 0;
       const size_t DEF_OUT_FREQ = 1000;
       out_freq = DEF_OUT_FREQ;
       //bool got_out_freq = false;
       use_pheno = false;
       std::ifstream in_file( config_filename );
       if (!in_file.is_open())
       {
           std::cout << "read_configfile : ";
           std::cout << "error : could not open " << config_filename << std::endl;
           return FILE_ERROR;
       }
        std::cout << "Reading input from " << config_filename << std::endl;
        std::string line;
        while (getline(in_file,line))
        {
            std::string buff;
            std::stringstream ss(line);
            std::vector<std::string> line_elements;
            while (ss >> buff)
            {
                line_elements.push_back(buff);
            }
            const size_t num_elements = line_elements.size();
            if ( num_elements == 0 )
            {
                continue;
            }
            if ( num_elements != 2 )
            {
              std::cout << "read_configfile() : line in " << config_filename <<
                         "has " << num_elements << "columns.  it should have either 0 or 2."
                         << std::endl;
              return FILE_ERROR;
            }
            if ( ( line_elements[ 0 ] != "geno_filename" ) &&
                 ( line_elements[ 0 ] != "out_filename" ) &&
                 ( line_elements[ 0 ] != "start_rep" ) &&
                 ( line_elements[ 0 ] != "num_reps" ) &&
                 ( line_elements[ 0 ] != "pheno_filename" ) && 
                 ( line_elements[ 0 ] != "MAF_cutoff" ) &&
                 ( line_elements[ 0 ] != "min_MAC"    ) && 
                 ( line_elements[ 0 ] != "out_freq"   ) &&
                 ( line_elements[ 0 ] != "min_call_rate" ) )
            {
                 std::cout << "error: read_configfile(): Invalid string " << line_elements[0]
                      << " found in first column of " << config_filename << std::endl;
                 return FILE_ERROR;	
            }

            if ( "geno_filename" == line_elements[ 0 ] )
            {
               got_geno_filename = true;
               geno_filename = line_elements[1];
            }
            else
            if ( "out_filename" == line_elements[ 0 ] )
            {
               got_out_filename = true;
               out_filename = line_elements[1];
            }
            else
            if ( "start_rep" == line_elements[ 0 ] )
            {
              try
              {
                 start_rep = std::stoull(line_elements[1]);
                 got_start_rep = true;
              }
              catch( std::exception & e)
              {
                in_file.close();
                std::cout << "Error parsing string " << line_elements[1]
                      << " for start_rep in " << config_filename << std::endl;
                return FILE_ERROR; 
              }
            }
            else
            if ( "num_reps" == line_elements[ 0 ] )
            {
              try
              {
                 num_reps = std::stoull(line_elements[1]);
                 got_num_reps = true;
              }
              catch( std::exception & e)
              {
                in_file.close();
                std::cout << "Error parsing string " << line_elements[1]
                      << " for num_reps in " << config_filename << std::endl;
                return FILE_ERROR;
              }
            }
            else
            if ( "pheno_filename" == line_elements[0] )
            {
               use_pheno = true;
               pheno_filename = line_elements[1];
            }
            else
            if ( "MAF_cutoff" == line_elements[0] )
            {
              try
              {
                 MAF_cutoff = std::stod(line_elements[1]);
                 got_MAF_cutoff = true;
              }
              catch( std::exception & e)
              {
                in_file.close();
                std::cout << "Error parsing string " << line_elements[1]
                      << " for MAF_cutoff in " << config_filename << std::endl;
                return FILE_ERROR;
              }
              if ( ( MAF_cutoff > 1.0 ) || 
                   ( MAF_cutoff < 0.0 ) )
              {
                in_file.close();
                std::cout << "Invalid value " << line_elements[1]
                      << " for MAF_cutoff in " << config_filename << std::endl;
                return FILE_ERROR;
              }
            }
            else
            if ( "min_MAC" == line_elements[0] )
            {
              //min_MAC is kept for compatibility with old config files but is unused
              std::cout << "min_MAC parameter is deprecated.  ";
              std::cout << "It will be ignored." << std::endl;
            }
            else
            if ( "out_freq" == line_elements[ 0 ] )
            {
              try
              {
                 out_freq = std::stoull(line_elements[1]);
                 //got_out_freq = true;
              }
              catch( std::exception & e)
              {
                in_file.close();
                std::cout << "Error parsing string " << line_elements[1]
                      << " for out_freq in " << config_filename << std::endl;
                return FILE_ERROR;
              }
            }
            else
            if ( "min_call_rate" == line_elements[0] )
            {
              try
              {
                 min_call_rate = std::stod(line_elements[1]);
              }
              catch( std::exception & e)
              {
                in_file.close();
                std::cout << "Error parsing string " << line_elements[1]
                      << " for min_call_rate in " << config_filename << std::endl;
                return FILE_ERROR;
              }
              if ( ( min_call_rate > 1.0 ) || 
                   ( min_call_rate < 0.0 ) )
              {
                in_file.close();
                std::cout << "Invalid value " << line_elements[1]
                      << " for min_call_rate in " << config_filename << std::endl;
                return FILE_ERROR;
              }
            }

         } //while


         in_file.close();
         if ( false == got_MAF_cutoff )
         {
            std::cout << "error: " << "MAF_cutoff " << "not found in " << config_filename
                      << std::endl;
            return FILE_ERROR;
         }

         if ( false == got_geno_filename )
         {
            std::cout << "error: " << "geno_filename " << "not found in " << config_filename 
                      << std::endl;          
            return FILE_ERROR;
         }
         if ( false == got_out_filename )
         {
            std::cout << "error: " << "out_filename " << "not found in " << config_filename 
                      << std::endl;
            return FILE_ERROR;
         }
         if ( false == got_start_rep ) 
         {
            std::cout << "error: " << "start_rep " << "not found in " << config_filename  
                      << std::endl;
            return FILE_ERROR;
         }
         if ( false == got_num_reps )  
         {               
            std::cout << "error: " << "num_reps " << "not found in " << config_filename   
                      << std::endl;
            return FILE_ERROR;
         }

        return NO_ERROR;
    } //read_configfile()


    int calc_above_min_MAF_c_vec(
            const MatrixXd & G_mat,
                  double min_MAF,
                  VectorXd & c_vec )
    {
        if ( ( min_MAF < 0.0 ) ||
             ( min_MAF > 1.0 ) )
        {
            std::cout << "calc_above_min_MAF : min_MAF "
                << "should be between 0.0 and 1.0" << std::endl;
            return ARG_ERROR;
        }
        VectorXd MAF_vec;
        calc_MAF_vec( G_mat, MAF_vec );
        c_vec = VectorXd::Zero( MAF_vec.rows() );
        for ( size_t row = 0; row < MAF_vec.rows(); ++row )
        {
            if ( ( MAF_vec( row ) >= min_MAF ) )
                // || ( abs( MAF_vec( row ) - min_MAF ) < EPS ) )
            {
                c_vec(row) = 1.0;
            }
        }

        return NO_ERROR;
    }

    void calc_MAF_vec( const MatrixXd & G_mat, VectorXd & MAF_vec )
    {
        const double SCALE_FACTOR = 1.0  / ( 2.0 * G_mat.cols() );
        const VectorXd sum_vector = G_mat.rowwise().sum();
        const VectorXd raw_freq_vector = SCALE_FACTOR * sum_vector;
        MAF_vec = raw_freq_vector;              
        for ( size_t index = 0; index < sum_vector.rows() ; ++index )
        {    
            if ( raw_freq_vector( index ) > 0.5 )
            {
                MAF_vec( index ) = 1.0 - raw_freq_vector( index );
            }
        } //for         
    } //calc_MAF_vec


    int linear_sim_Y_vec_nocovar(
                    size_t seed_it,
                    size_t NUM_CASES,
                    VectorXd & Y_vec )             
    {
        Y_vec = VectorXd::Zero( NUM_CASES );
        std::mt19937_64 mt_engine;   //mersenne twister
        size_t seed = 123456789 + 2307 * seed_it;
        mt_engine.seed( seed );                   
        //std normal - note that
        //norm_dis takes mean and standard dev, not variance
        std::normal_distribution<double> norm_dis(0.0, 1.0 );

        for ( size_t subject = 0; subject < NUM_CASES; ++subject )
        {
            const double r_normal = norm_dis( mt_engine );
            Y_vec( subject ) = r_normal;
        }
        return NO_ERROR;
    } //linear_sim_Y_vec_nocovar

    int resample_threshold_linear_nocovar(
            const std::vector< MatrixXd > & geno_vec,
            const std::string & outfilename,
            size_t start_rep,
            size_t num_reps,
            size_t out_freq   )
    {
        //std::cout << "geno_vec.size() == " << geno_vec.size() << std::endl;
        //derived from resample_Burden_threshold_linear_nocovar 
        //method in generalized code        
        
        //std::cout << 
        //        "warning : resample_threshold_linear_nocovar should be used for no covariates "
        //        << "and no phenotype. If the simulation has phenotype this should not be used." << std::endl;
        
        if ( start_rep < 0 )
        {
            std::cout << 
                "resample_threshold_linear_nocovar error : " <<
                "start_rep is negative." << std::endl;
            return ARG_ERROR;
        }
        if ( num_reps < 0 )
        {
            std::cout << 
            "resample_threshold_linear_nocovar error : " <<
                "num_reps is negative." << std::endl;
            return ARG_ERROR;
        }
        if ( out_freq <= 0 )
        {
            std::cout <<
            "resample_threshold_linear_nocovar error : " <<
                "out_freq == " << out_freq << std::endl;
            return ARG_ERROR;
        }
        std::vector<double> V_val_vec;
        std::vector< VectorXd > sum_G_vxd_vec;  
        if ( geno_vec.size() == 0 )
        {
           std::cout << 
                "resample_threshold_linear_nocovar : " <<
                " geno_vec is empty " << std::endl;
            return NO_ERROR;
        }

        std::vector< MatrixXd > filtered_geno_vec;
        for ( const auto & G_mat : geno_vec )
        {
			//std::cout << "G-mat.rows() == " << G_mat.rows() << std::endl;
			//std::cout << "G-mat.cols() == " << G_mat.cols() << std::endl;
           if ( G_mat.cols() == 0 )
           {
               continue;
           }
		/*
           size_t non_zero_geno_count = 0;
           for ( size_t row = 0; row < G_mat.rows(); ++row )
           {
           		if ( G_mat( row, 0 ) > 0.5 )
                {
                    ++non_zero_geno_count;
                }
           }
           if ( non_zero_geno_count < min_mac )
           {
                    continue;
           }
         */ 
           filtered_geno_vec.push_back( G_mat );
        } //for ( const auto & G_mat : geno_vec )
        if ( filtered_geno_vec.size() == 0 )
        {
           std::cout << "Error: resample_threshold_linear_nocovar() " << std::endl
                     << "All SNPs had zero length (no subjects)." << std::endl;
           return ARG_ERROR;
        }
        const size_t NUM_SUBJECTS = filtered_geno_vec[ 0 ].rows();
std::cout << std::endl << "Number of SNPs == " << filtered_geno_vec.size() << std::endl; 
        for ( auto & G_mat : filtered_geno_vec )
        {
            if ( G_mat.rows() != NUM_SUBJECTS )
            {
                std::cout << 
                    "resample_threshold_linear_nocovar error : " <<
                    " NUM_SUBJECTS != G_mat.rows() for a G_mat in geno_vec"
                    << std::endl;
                    return DIM_ERROR;
            } //if
            const VectorXd sum_G_vxd = G_mat.rowwise().sum();
            V_val_vec.push_back( sum_G_vxd.dot( sum_G_vxd ) );
            sum_G_vxd_vec.push_back( sum_G_vxd );
        } //for
        std::ofstream outfile( outfilename );
        std::cout << "resample_threshold_linear_nocovar : begin iter loop" <<
                std::endl;
        for ( size_t iter = start_rep; iter < (num_reps + start_rep); ++iter )
        {
                if ( ( iter % out_freq ) == 0 )
                {
                    std::cout << "iter == " << iter << std::endl;
                }
                double max_this_iter_Q_val = -100.0;
                double max_this_iter_Q_val_alt = -100.0;
                VectorXd Y_star_vec;
                
                //step 1: create simulated Y vector
                int lin_reg_rv = linear_sim_Y_vec_nocovar(
                    iter,
                    NUM_SUBJECTS,
                    Y_star_vec );
                if ( lin_reg_rv )
                {
                    std::cout << "resample_threshold_linear_nocovar error : " <<
                        "linear_sim_Y_vec_nocovar returned non-zero value == " <<
                        lin_reg_rv << " line : " <<  __LINE__ <<
                        "\n iter : " << iter <<
                        std::endl;
                    return SIM_ERROR;
                }
                
            size_t geno_index = 0;
            for ( const auto & G_mat : filtered_geno_vec )
            {
                const VectorXd sum_G_vxd = sum_G_vxd_vec[ geno_index ];
                const double V_val = V_val_vec[ geno_index ];
                const double U_val = Y_star_vec.dot( sum_G_vxd );
                const double Q_val = ( U_val * U_val ) / V_val;
                if ( Q_val > max_this_iter_Q_val )
                {
                   max_this_iter_Q_val = Q_val;
                }
                geno_index++;
            } //for G_mat
            outfile << max_this_iter_Q_val;
            outfile << "\n";
        } //for iter loop
        outfile.close();
        return NO_ERROR;
    } //resample_threshold_linear_nocovar

    int read_geno_file( const std::string & filename,
                    std::vector< MatrixXd > &  geno_mat_vec,
                    std::vector< std::string > & gene_names_vec,
                    std::vector< subject_data_type > & subject_data_vec_missing_pheno,
                    double min_call_rate 
                    )
    {
        std::vector< std::vector< subject_data_type > >
                                         subject_data_vectors;
        const int NA_geno_value = -10; //should be negative
        if ( ( min_call_rate > 1.0 ) || 
             ( min_call_rate < 0.0 ) )
        {
           std::cout << "error: read_geno_file() : minimum call rate set to " <<
                 min_call_rate << std::endl; 
           return ARG_ERROR;
        }
        std::ifstream in_file( filename );
        if (!in_file.is_open())
        {
            std::cout << "error in read_geno_file() : file " 
                     << filename << " could not be opened." <<
                    std::endl;
            return FILE_ERROR;
        }
        std::string line;
        while (std::getline(in_file,line)) 
        {
            std::string buff;
            std::stringstream ss(line);
            std::vector <std::string> line_elements;
            while (ss >> buff)   
            {              
                const std::string element = buff;
                line_elements.push_back(element);           
            }
            if ( (line_elements.size() < 2 ) )
            {  
                std::cout << "error : each line of genotype file should have "
                    << "at least 2 columns"
                    << std::endl;
                in_file.close();
                return FILE_ERROR;
            }
            const std::string snp_name = line_elements[ 0 ];
       //     line_elements.erase( line_elements.begin() );
            gene_names_vec.push_back( snp_name );
            
            std::vector<double> db_vector;
            //for ( const auto & element : line_elements )
			for ( size_t i = 1; i < line_elements.size(); ++i )
            {
					std::string element = line_elements[ i ]; 
                    int current_int;
                    try
                    {
                        current_int = stoi( element );
                        if ( ( current_int != 0 ) &&
                             ( current_int != 1 ) &&
                            ( current_int != 2 ) )
                        {
                            std::cout <<
                            "read_geno_file : error : invalid int read"
                            << " == " << current_int << std::endl;
                            in_file.close();
                            return FILE_ERROR;
                        } //if
                        db_vector.push_back( current_int );
                    } //try
                    catch( std::exception & e )
                    {
                        if ( element.compare( "NA" ) )
                        {
                            std::cout <<
                            "read_geno_file : error : invalid int read"
                            << " == " << current_int << std::endl;
                            in_file.close();
                            return FILE_ERROR;
                        } //if
                        db_vector.push_back( NA_geno_value );
                    } //catch
            } //for

            MatrixXd my_geno_mat = MatrixXd::Zero( 1, (line_elements.size() - 1) );
            for ( size_t my_index = 0; my_index < (line_elements.size() - 1)   ; ++my_index )
            {
                my_geno_mat(0, my_index) = db_vector[ my_index ];
            }
            geno_mat_vec.push_back( my_geno_mat );
       } //while
     in_file.close();


for ( auto & current_mat : geno_mat_vec )
{
    std::vector< bool > call_rate_check_passed;
    for ( size_t row = 0; row < current_mat.rows(); ++row )
    {
        size_t missing_count = 0;
        for ( size_t my_case = 0; my_case < current_mat.cols(); ++my_case )
        {
            if  ( current_mat( row, my_case ) < ( NA_geno_value / 2.0 ) )
            {
                missing_count++;
            } //if
        } //for case
        const size_t MAX_ALLOWED_MISSING = round( ( 1.0 - min_call_rate ) * current_mat.cols() );
        
//std::cout << "MAX_ALLOWED_MISSING == " << MAX_ALLOWED_MISSING << std::endl;
//std::cout << "missing_count == " << missing_count << std::endl;
		if ( missing_count > MAX_ALLOWED_MISSING )
        {
            call_rate_check_passed.push_back( false );
        }
        else
        {
            call_rate_check_passed.push_back( true );
        }

      int my_int;
       bool all_equal = true;
       for ( size_t  my_case = 0; my_case < current_mat.cols(); ++my_case )            
       {
             if ( my_case == 0 )
             {
                my_int = int( current_mat( row, my_case ) );
             }                   
             else
             {
               if ( my_int != int( current_mat( row, my_case ) ) )
               {
                   all_equal = false;
                   break;
               }                         
             }     

       }
       if ( all_equal == true )
        {
                call_rate_check_passed.pop_back();
            call_rate_check_passed.push_back( false );
        }


    } //for row
    size_t number_good_rows = 0;
    for ( const auto & val : call_rate_check_passed )
    {
        if ( val )
        {
            number_good_rows++;
        }
    
    }
    MatrixXd new_G_mat = MatrixXd::Zero( number_good_rows, current_mat.cols() );
    for ( size_t coor = 0, good_index = 0;
                        coor < current_mat.rows();
                        ++coor )
    {
        if ( call_rate_check_passed[ coor ] == false )
        {
            //this snp will be removed
            continue;
        }
        new_G_mat.block( good_index, 0, 1, current_mat.cols() ) =
                current_mat.row( coor );
        good_index++;
    }
    current_mat = new_G_mat;
} //for matrix

    for ( auto & current_mat : geno_mat_vec )
    {
        std::vector< subject_data_type > current_subject_data_type_vec;
        for ( size_t my_case = 0; my_case < current_mat.cols(); ++my_case )
        {
            bool missing_in_this_case = false;
            std::vector<int> data_for_snp;
            for ( size_t my_coor = 0; my_coor < current_mat.rows(); ++my_coor )
            {
                if ( current_mat( my_coor, my_case ) < ( NA_geno_value / 2.0 ) )
                {
                    data_for_snp.push_back( 0 );
                   // missing_in_this_case = true;
                   // break;
                }
                else
                {
                    data_for_snp.push_back( 1 );
                }
            } //for my_coor
            size_t sum = 0;
            for ( auto & val : data_for_snp )
            {
                sum += val;
            }
            if ( sum == current_mat.rows() )
            {
                missing_in_this_case = false;
            }
            //else
            //if ( sum == 0 )
            //{
            //    missing_in_this_case = true;
            //
            //}
            else
            {
                //impute the missing
                for ( size_t row = 0; row < current_mat.rows(); ++row )
                {
                    if ( data_for_snp[ row ] )
                    {
                        continue;
                    }
                    //else impute
                    size_t valid_in_this_row = 0;
                    size_t valid_sum = 0;
                    for ( size_t my_subj = 0;
                                    my_subj < current_mat.cols();
                                ++my_subj )
                    {
                        if ( current_mat( row, my_subj ) > ( NA_geno_value / 2.0 ) )
                        {
                            valid_in_this_row++;
                            valid_sum += current_mat( row, my_subj );
                        }
                    } //for my_subj
                    if ( ! valid_in_this_row )
                    {
                        //no subject has valid data for this snp
                        missing_in_this_case = true;
                    } 
                    else
                    {
                        //can impute
                        missing_in_this_case = false;
                        current_mat( row, my_case ) = round( ( valid_sum * 1.0 ) / valid_in_this_row ); 
                    }                   
                } //for row      
            } //else
            current_subject_data_type_vec.push_back( FULL_DATA ); //don't delete subj for miss geno
            //if ( ! missing_in_this_case ) ||  
            //{
            //    current_subject_data_type_vec.push_back( FULL_DATA );
            //}
            //else
            //{
            //    current_subject_data_type_vec.push_back( MISSING_DATA );
            //}
        } //for my_case
        subject_data_vectors.push_back( current_subject_data_type_vec );
    } //for current_mat
    std::vector< MatrixXd > new_geno_mat_vec;

    for ( size_t index = 0; index < geno_mat_vec.size(); ++index )
    {
        const MatrixXd orig_mat = geno_mat_vec[ index ];
        const size_t NUM_COOR = orig_mat.rows();
        size_t NUM_FULL_DATA_SUBJECTS = 0;
        std::vector< subject_data_type > my_subject_data_vec = 
            subject_data_vectors[ index ];
        //if this simulation is without pheno
        //subject_data_vec_missing_pheno must be filled
        if ( subject_data_vec_missing_pheno.size() == 0 )
        {
             for ( size_t subj = 0; subj < my_subject_data_vec.size(); ++subj )
             {
                 subject_data_vec_missing_pheno.push_back( FULL_DATA );
             }
        }
        for ( size_t subj = 0; subj < my_subject_data_vec.size(); ++subj )
        {
            if ( ( my_subject_data_vec[ subj ] == FULL_DATA ) && 
                 ( subject_data_vec_missing_pheno[ subj ]  == FULL_DATA ) )
            {
                NUM_FULL_DATA_SUBJECTS++;
            }
        }
        MatrixXd full_data_only_mat = 
            MatrixXd::Zero( NUM_COOR, NUM_FULL_DATA_SUBJECTS );
        
        for ( size_t subj = 0, known_index = 0; 
                     subj < orig_mat.cols(); 
                   ++subj )
        {
            if ( my_subject_data_vec[ subj ] == MISSING_DATA )
            {   //missing geno data
                continue;
            }
            //else
            if ( subject_data_vec_missing_pheno[ subj ] ==
                    MISSING_DATA )
            {
                //missing pheno data
                continue;
            }
            //else
            full_data_only_mat.block( 0, known_index, NUM_COOR, 1 ) =
                orig_mat.col( subj );
            known_index++;
        }
        new_geno_mat_vec.push_back( full_data_only_mat );
    }
     geno_mat_vec = new_geno_mat_vec;
     return NO_ERROR;
    } //read_geno_file


int perform_linear_reg(const MatrixXd & X_mat,
                   const VectorXd & Y_vec,
                   VectorXd & beta_vec,
                   VectorXd & se_vec,
                double & sigma_sqr, //denominator is N - P (used for SE)
                double & sigma_sqr_alt //denominator is N (not used for SE)
                   )
{
    const size_t NUM_CASES = X_mat.cols(); //n
    const size_t NUM_COOR = X_mat.rows(); // p
    //check dimensions
    if ( Y_vec.rows() != NUM_CASES )
    {
        std::cout << "Error: perform_linear_reg : "
            << "Y_vec.rows != X_mat.cols() " << std::endl;
        return DIM_ERROR;
    }
    MatrixXd sum_prod_mat = MatrixXd::Zero(NUM_COOR, NUM_COOR);

    for (size_t mat_index = 0; mat_index < NUM_CASES; mat_index++)
    {
        sum_prod_mat += X_mat.col( mat_index ) *
                        ( ( X_mat.col( mat_index ) ).transpose() );
    }
    FullPivLU<MatrixXd> lu = sum_prod_mat.fullPivLu();
    if ( ! lu.isInvertible() )
    {
        std::cout << "Error: sum_prod_mat Matrix not invertible." << std::endl;
        return ARG_ERROR;
    }
    MatrixXd inv_sum_prod_mat = sum_prod_mat.inverse();
    MatrixXd right_side_vec = MatrixXd::Zero(NUM_COOR, 1);
    for (size_t mat_index = 0; mat_index < NUM_CASES; mat_index++)
    {
        right_side_vec += Y_vec(mat_index) * X_mat.col( mat_index );
    }
    beta_vec = inv_sum_prod_mat * right_side_vec;
    double numerator = 0.0;
    for (size_t mat_index = 0; mat_index < NUM_CASES; mat_index++)
    {
        double yi =  Y_vec( mat_index );
        double prod = beta_vec.transpose() * X_mat.col(mat_index);
        double diff = yi - prod;
        numerator += diff*diff;
    }
    const double DENOM = NUM_CASES - NUM_COOR;
    sigma_sqr = numerator / DENOM;
    sigma_sqr_alt = numerator / NUM_CASES;
    MatrixXd cov_mat = sigma_sqr * inv_sum_prod_mat;
    se_vec = VectorXd::Zero(NUM_COOR, 1);

    for (size_t index = 0; index < NUM_COOR; index++)
    {
       se_vec( index ) = sqrt( cov_mat( index, index ) );
    }
    return NO_ERROR;
}


    void calc_U_vec_linear(
            double sigma_sqr,
            const VectorXd & gamma,
            const MatrixXd & G_mat,
            const MatrixXd & X_mat,  
            const VectorXd & Y_vec,
                  VectorXd & U_vec )   
    {

        const size_t NUM_COOR_G = G_mat.rows();
        const size_t NUM_COOR_X = X_mat.rows();
        const size_t NUM_CASES = Y_vec.rows();

        U_vec = VectorXd::Zero( NUM_COOR_G );
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            const double Y_val = Y_vec( index );       
            const VectorXd G_vec = G_mat.col( index );
            const VectorXd X_vec = X_mat.col( index );
            const double resid = Y_val - gamma.dot( X_vec );
            const VectorXd current_vec = resid * G_vec;
            U_vec += current_vec;
        }
        //scale by sigma_sqr :
        U_vec /= sigma_sqr;

    } //calc_U_vec_linear

    int calc_V_mat_linear(
            double sigma_sqr,
            const MatrixXd & G_mat,
            const MatrixXd & X_mat,
                  MatrixXd & V_mat )
    {
        const size_t NUM_COOR_G = G_mat.rows();
        const size_t NUM_COOR_X  = X_mat.rows();
        const size_t NUM_CASES = G_mat.cols();
        if ( G_mat.cols() != X_mat.cols() )
        {
            std::cout << "calc_V_mat_linear : G_mat and X_mat "
                << "have differing numbers of columns.\n";
            return 2;
        }
        const double SCALE_FACTOR = 1.0 / sigma_sqr;
        MatrixXd sum_G_sqr_mat =
            MatrixXd::Zero( NUM_COOR_G, NUM_COOR_G );
        MatrixXd sum_G_X_mat =
            MatrixXd::Zero( NUM_COOR_G, NUM_COOR_X );
        MatrixXd sum_X_sqr_mat =
            MatrixXd::Zero( NUM_COOR_X, NUM_COOR_X );

        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            const VectorXd G_vec = G_mat.col( index );
            const VectorXd X_vec = X_mat.col( index );

            sum_G_sqr_mat += G_vec * (G_vec.transpose());

            sum_G_X_mat += G_vec * ( X_vec.transpose() );

            sum_X_sqr_mat += X_vec * ( X_vec.transpose() );
        }
        FullPivLU<MatrixXd> lu = sum_X_sqr_mat.fullPivLu();
        if ( ! lu.isInvertible() )
        {
            std::cout <<
                "Error: calc_V_mat_linear : matrix not invertible." <<
                    std::endl;
            return ARG_ERROR;
        }

        const MatrixXd inv_X_sqr_mat = sum_X_sqr_mat.inverse();

        const MatrixXd second_term_mat =
            sum_G_X_mat * inv_X_sqr_mat * ( sum_G_X_mat.transpose() );
        const MatrixXd unscaled_result_mat =
            sum_G_sqr_mat - second_term_mat;
        V_mat = SCALE_FACTOR * unscaled_result_mat;
        return NO_ERROR;
    }
    int calc_Burden_Q(
            const VectorXd & c_vec,     
            const VectorXd & U_vec,
            const MatrixXd & V_mat,
                  double & Q_val )
    {
        //this is the square of the Burden T value
        //this function avoids an unnecessary call to sqrt
        const size_t NUM_COOR = c_vec.rows();                  
        if ( U_vec.rows() != NUM_COOR )
        {
            std::cout << "calc_Burden_Q : error : c_vec and U_vec have"
                <<  " differing numbers of rows." << std::endl;
            std::cout << "c_vec.rows() == " << c_vec.rows() << std::endl;
            std::cout << "U_vec.rows() == " << U_vec.rows() << std::endl;
            return ARG_ERROR;
        }
        if ( V_mat.rows() != NUM_COOR )
        {
            std::cout << "calc_Burden_Q : error : c_vec and V_mat have"
                <<  " differing numbers of rows." << std::endl;
            return ARG_ERROR;
        }
        if ( V_mat.rows() != V_mat.cols() )
        {
            std::cout << "calc_Burden_Q : error : V_mat is not a "
                <<  "square matrix." << std::endl;
            return ARG_ERROR;
        }
        const double numerator = pow( c_vec.transpose() * U_vec, 2.0 );
        const double denominator  = c_vec.transpose() * V_mat * c_vec;
        Q_val = numerator / denominator;
        return NO_ERROR;
    }

int read_pheno_data( std::string infilename,
                MatrixXd & X_mat,
                MatrixXd & Y_col_vec,
                std::vector< subject_data_type > & subject_data_vec                
				)
{
    std::ifstream in_file( infilename.c_str() );
    if (!in_file.is_open())
    {
        std::cout << "read_pheno_data : ";
        std::cout << "error : could not open " << infilename << std::endl;
        return FILE_ERROR;
    }
    else
    {
        std::cout << "Reading input from " << infilename << std::endl;
        std::string line;
        size_t num_X_cols = 0;
        size_t num_rows = 0;
        size_t full_data_count = 0;
        while (getline(in_file,line))
        {
            std::string buff;
            std::stringstream ss(line);
            std::vector <double> line_elements;
            std::vector <std::string> line_strings;
            while (ss >> buff)
            {
                double element = atof(buff.c_str());
                line_elements.push_back(element);
                line_strings.push_back( buff );
            }
            if ( line_strings[0].compare( "NA" ) )
            {
                subject_data_vec.push_back( FULL_DATA );
                ++full_data_count;
            }
            else
            {
                subject_data_vec.push_back( MISSING_DATA );
            }

            if (line_elements.size() > 0)
            {
                num_rows++;
                if (num_X_cols == 0)
                {
                    //first time
                    num_X_cols = line_elements.size();
                    X_mat.resize(num_rows, num_X_cols);
                    Y_col_vec.resize(num_rows, 1);
                }
                else
                {
                    X_mat.conservativeResize(num_rows, NoChange);
                    Y_col_vec.conservativeResize(num_rows, NoChange);
                }
                Y_col_vec(num_rows - 1, 0) = line_elements[0];
                for (size_t index = 1; index < line_elements.size(); index++)
                {             
                    X_mat( num_rows - 1, index) = line_elements[index];
                }
            } //if line_elements.size() ...
        }   //while
        in_file.close();
        //add extra column of ones to X_mat
        for (size_t index = 0; index < X_mat.rows(); index++)
        {
            X_mat(index, 0) = 1.0;
        }
        //transpose X_mat
        X_mat.transposeInPlace();
        //now remove subjects from Y_col_vec that have missing phenotype data
        VectorXd full_Y_vec = VectorXd::Zero( full_data_count );
        for ( size_t index = 0, full_data_index = 0;
                     index < subject_data_vec.size(); ++index )
        {
            if ( subject_data_vec[ index ] == MISSING_DATA )
            {
                continue;
            }
            full_Y_vec( full_data_index ) = Y_col_vec( index );
            ++full_data_index;
        }
        Y_col_vec = full_Y_vec;

        //now remove subjects from X_mat that have missing
        //phenotype data
        MatrixXd full_X_mat = MatrixXd::Zero( X_mat.rows(), full_data_count );
        for ( size_t index = 0, full_data_index = 0;
                     index < subject_data_vec.size(); ++index )
        {
            if ( subject_data_vec[ index ] == MISSING_DATA )
            {
                continue;
            }
            full_X_mat.block( 0, full_data_index, X_mat.rows(), 1 ) =
                X_mat.block( 0, index, X_mat.rows(), 1 );
            ++full_data_index;
        }
        X_mat = full_X_mat;
        std::cout << "done reading phenotype data. " << std::endl;    
        return NO_ERROR;
    } //else  
} //read_pheno_data

    int resample_Burden_threshold_linear(
            const MatrixXd & X_mat,
            const VectorXd & Y_vec,
            const std::vector< MatrixXd > & geno_vec,
            const std::string & outfilename,
            size_t start_rep,
            size_t num_reps,
			size_t out_freq )
    {
        if ( start_rep < 0 )
        {
            std::cout << "resample_Burden_threshold_linear error : " <<
                "start_rep is negative." << std::endl;
            return ARG_ERROR;
        }
        if ( num_reps < 0 )
        {
            std::cout << "resample_Burden_threshold_linear error : " <<
                "num_reps is negative." << std::endl;
            return ARG_ERROR;
        }
        const size_t NUM_SUBJECTS = Y_vec.rows();
        if ( X_mat.cols() != NUM_SUBJECTS )
        {
            std::cout << "resample_Burden_threshold_linear error : " <<
                " Y_vec.rows() != X_mat.cols() " << std::endl;
            return DIM_ERROR;
        }
        if ( geno_vec.size() == 0 )
        {
           std::cout << "resample_Burden_threshold_linear : " <<
                " geno_vec is empty " << std::endl;
            return DIM_ERROR;
        }
        for ( auto & G_mat : geno_vec )
        {
            if ( G_mat.rows() != NUM_SUBJECTS )
            {
                std::cout << "resample_Burden_threshold_linear error : " <<
                    " Y_vec.rows() != G_mat.rows() for a G_mat in geno_vec"
                    << std::endl;
                    return DIM_ERROR;
            } //if
        } //for
        VectorXd lr_gamma_hat, lr_gamma_hat_se;
        double lr_sigma_sqr, lr_sigma_sqr_alt;
        int lin_reg_rv = perform_linear_reg(
            X_mat, Y_vec, lr_gamma_hat,
            lr_gamma_hat_se, lr_sigma_sqr, lr_sigma_sqr_alt );
        if ( lin_reg_rv != NO_ERROR)
        {
            std::cout << "resample_Burden_threshold_linear error : " <<
                "linear regression returned non-zero value == " <<
                lin_reg_rv << "\tline == " << __LINE__ << std::endl;
            return lin_reg_rv;
        }

        std::vector< MatrixXd > filtered_geno_vec;
        for ( const auto & G_mat : geno_vec )
        {
           //size_t non_zero_geno_count = 0;
			//std::cout << "G_mat.rows() == " << G_mat.rows() << std::endl;
			//std::cout << "G_mat.cols() == " << G_mat.cols() << std::endl;
           if ( G_mat.cols() == 0 )
           { //skip snps that did not pass MAF check
              continue;
           }
		/*
           for ( size_t row = 0; row < G_mat.rows(); ++row )
           {
				
                if ( G_mat( row, 0 ) > 0.5 )
                {
                    ++non_zero_geno_count;
                }
           }
           if ( non_zero_geno_count < min_mac )
           {
                    continue;
           }
		*/

           filtered_geno_vec.push_back( G_mat.transpose() );
        } //for ( const auto & G_mat : geno_vec )
		std::cout << std::endl << "Number of SNPs == " << filtered_geno_vec.size() 
				  << std::endl;
        //calculate Burden c vectors for each gene
        //These will be reused during the iter loop
        std::vector< VectorXd > Burden_c_vecs;
        for ( const auto & G_mat : filtered_geno_vec )
        {
            const VectorXd c_vec = VectorXd::Ones( G_mat.rows() );
            Burden_c_vecs.push_back( c_vec );
        } //for G_mat
        std::ofstream outfile( outfilename );
        std::cout << "resample_Burden_threshold_linear : begin iter loop" <<
                std::endl;
        for ( size_t iter = start_rep; iter < (num_reps + start_rep); ++iter )
        {
                if ( ( iter % out_freq ) == 0 )
                {
                    std::cout << "iter == " << iter << std::endl;
                }
                double max_this_iter_Q_val = -100.0;
                VectorXd Y_star_vec;
                //step 1: create simulated Y vector
                int lin_reg_rv = linear_sim_Y_vec_from_null(
                    iter,
                    X_mat,
                    lr_gamma_hat, //from linear regression
                    lr_sigma_sqr_alt, //from linear regression
                    Y_star_vec );
                if ( lin_reg_rv != NO_ERROR )
                {
                    std::cout << "resample_Burden_threshold_linear error : " <<
                        "linear_sim_Y_vec_from_null returned non-zero value == " <<
                        lin_reg_rv << " line : " <<  __LINE__ <<
                        "\n iter : " << iter <<
                        "\n lr_gamma_hat == \n" << lr_gamma_hat <<
                        "\n lr_sigma_sqr == " << lr_sigma_sqr  <<
                        std::endl;
                    return lin_reg_rv;
                }
               //step 2) calculate restricted MLEs
                VectorXd lr_gamma_star_hat;
                VectorXd lr_gamma_star_hat_se;
                double lr_sigma_sqr_star, lr_sigma_sqr_star_alt;
                lin_reg_rv = perform_linear_reg(
                    X_mat, Y_star_vec, lr_gamma_star_hat,
                    lr_gamma_star_hat_se, lr_sigma_sqr_star,
                    lr_sigma_sqr_star_alt );
                if ( lin_reg_rv != NO_ERROR )
                {
                    std::cout << "resample_Burden_threshold_linear error : " <<
                        "linear regression returned non-zero value == " <<
                        lin_reg_rv << "line == " << __LINE__ <<
                        " iter == " << iter << std::endl;
                    return lin_reg_rv;
                }
            size_t geno_index = 0;
            for ( const auto & G_mat : filtered_geno_vec )
            {
                const VectorXd c_vec = Burden_c_vecs[ geno_index ];
                //std::cout << "c_vec == " << c_vec << std::endl;
                VectorXd U_vec;
                calc_U_vec_linear(
                            lr_sigma_sqr_star_alt,
                            lr_gamma_star_hat,
                            G_mat,
                            X_mat,
                            Y_star_vec,
                            U_vec );
                MatrixXd V_mat;                
                const int calc_V_rv = 
                    calc_V_mat_linear( lr_sigma_sqr_star_alt,
                                       G_mat,
                                       X_mat,
                                       V_mat );
                if ( calc_V_rv != NO_ERROR )
                {
                    std::cout << "resample_Burden_threshold_linear : " <<
                       " calc_V_mat_linear returned " << "non-zero "<<
                       "value == " << calc_V_rv << " iter == " << iter << 
                        std::endl;
                    return calc_V_rv;
                }
                double Q_val;
                const int calc_Burden_Q_rv =
                    calc_Burden_Q( c_vec, U_vec, V_mat, Q_val );
                if ( calc_Burden_Q_rv != NO_ERROR)
                {
                    std::cout << "resample_Burden_threshold_linear : " <<
                       " calc_Burden_Q returned " << "non-zero "<< 
                       "value == " << calc_Burden_Q_rv  << " iter == " << iter <<
                        std::endl;  
                    return calc_Burden_Q_rv;
                }
                if ( Q_val > max_this_iter_Q_val )
                {
                   max_this_iter_Q_val = Q_val;
                }
                geno_index++;
            } //for G_mat
            outfile << max_this_iter_Q_val;
            outfile << "\n";
        } //for iter loop
        outfile.close();
        //find threshold
        return NO_ERROR;
    } //resample_Burden_threshold_linear


    int linear_sim_Y_vec_from_null( 
                  size_t seed_it,
            const MatrixXd & X_mat,
            const VectorXd & gamma, //from linear regression
                  double sigma_sqr, //from linear regression
                  VectorXd & Y_vec )
    {
        const size_t NUM_X_COOR = X_mat.rows();
        if ( gamma.rows() != NUM_X_COOR )
        {
            std::cout << "linear_sim_Y_vec_from_null error : " <<
                "X_mat and gamma do not have same number of rows. " << std::endl;
            return DIM_ERROR;
        }
        if ( sigma_sqr < 0.0 )
        {
            std::cout << "linear_sim_Y_vec_from_null error : " <<
                "sigma_sqr is negative" << std::endl;
            return ARG_ERROR;
        }            
        Y_vec = VectorXd::Zero( X_mat.cols() );
        std::mt19937_64 mt_engine;   //mersenne twister
        size_t seed = 123456789 + 2307 * seed_it;
        mt_engine.seed( seed );          
        std::normal_distribution<double> norm_dis(0.0, sqrt(sigma_sqr) );
        
        for ( size_t subject = 0; subject < X_mat.cols(); ++subject )
        {
            const VectorXd X_vec = X_mat.col( subject );
            const double r_normal = norm_dis( mt_engine );
            Y_vec( subject ) = gamma.dot( X_vec ) + r_normal;      
        }        
        return NO_ERROR;
    } //linear_sim_Y_vec_from_null

} //namespace single_var
