#ifndef SINGLE_VAR_H
#define SINGLE_VAR_H
#include <vector>                           
#include <string>                  
#include <iostream>                  
#include <fstream>                   
#include <eigen3/Eigen/Dense>  
#include <cfloat>
#include <random>                  
#include <exception>
using namespace Eigen;

namespace single_var
{

    //for phenotype/covariates       
    enum subject_data_type
    {               
        FULL_DATA,
        MISSING_DATA
    };

    //methods
    int linear_sim_Y_vec_nocovar(
        size_t seed_it,
        size_t NUM_CASES,
        VectorXd & Y_vec );

    int resample_threshold_linear_nocovar(
            const std::vector< MatrixXd > & geno_vec,
            const std::string & outfilename,
                  size_t start_rep = 0,
                  size_t num_reps = 10000,
                  size_t out_freq = 1000 );


    int read_geno_file( const std::string & filename,
                    std::vector< MatrixXd > &  geno_mat_vec,                        
                    std::vector< std::string > & gene_names_vec,
                    std::vector< subject_data_type > & subject_data_vec_missing_pheno,
                    double min_call_rate );


    int perform_linear_reg(const MatrixXd & X_mat,
                   const VectorXd & Y_vec,
                   VectorXd & beta_vec,
                   VectorXd & se_vec,
                   double & sigma_sqr,
                   double & sigma_sqr_alt );

    void calc_U_vec_linear(    
            double sigma_sqr,
            const VectorXd & gamma,
            const MatrixXd & G_mat,
            const MatrixXd & X_mat,
            const VectorXd & Y_vec,
                  VectorXd & U_vec );


    int calc_V_mat_linear(     
            double sigma_sqr,
            const MatrixXd & G_mat,
            const MatrixXd & X_mat,
                  MatrixXd & V_mat );

    int calc_Burden_Q(         
            const VectorXd & c_vec,
            const VectorXd & U_vec,
            const MatrixXd & V_mat,
                  double & Q_val  );

     int read_pheno_data( std::string infilename,
                MatrixXd & X_mat,
                MatrixXd & Y_col_vec,
                std::vector< subject_data_type > & subject_data_vec
                );



     int resample_Burden_threshold_linear(
            const MatrixXd & X_mat,
            const VectorXd & Y_vec,
            const std::vector< MatrixXd > & geno_vec,
            const std::string & outfilename,
                size_t start_rep = 0,
                size_t num_reps = 10000,
                size_t out_freq = 1000 );

    int linear_sim_Y_vec_from_null( 
                  size_t seed_it,
            const MatrixXd & X_mat,
            const VectorXd & gamma, //from linear regression
                  double sigma_sqr, //from linear regression
                  VectorXd & Y_vec );

    int calc_above_min_MAF_c_vec(
            const MatrixXd & G_mat,
                  double min_MAF,
                  VectorXd & c_vec );


    void calc_MAF_vec(
                const MatrixXd & G_mat,
                      VectorXd & MAF_vec );



    int read_configfile(
        const std::string & config_filename,
              std::string & geno_filename,
              std::string & out_filename,
                   size_t & start_rep,
                   size_t & num_reps,  
                   double & MAF_cutoff,            
                 bool & use_pheno,
                 std::string & pheno_filename,
                   size_t & out_freq,
                   double & min_call_rate  );


               
    //return values
    const int NO_ERROR  = 0;
    const int DIM_ERROR = 2;
    const int ARG_ERROR = 4;
    const int SIM_ERROR = 8;
    const int FILE_ERROR = 16;

}
#endif

